#!/usr/bin/env python

from typing import Optional
import argparse
import os
import os.path
import sys

from unmixer.ui import UnmixerUI
from unmixer.util import expand_path, is_isolated_track


def parse_args(args: list[str]) -> argparse.Namespace:
    parser = argparse.ArgumentParser(description='create and explore isolated tracks from music files')
    parser.add_argument('music_file_or_track_dir', nargs='?',
                        help='path to the file to unmix, or path to a directory containing isolated tracks')
    return parser.parse_args(args)


def main(args: Optional[list[str]] = None) -> None:
    if args is None:
        args = sys.argv[1:]
    
    config = parse_args(args)
    input_dir_path = None
    song_path = None
    
    if path := config.music_file_or_track_dir:
        path = expand_path(path)
        if not os.path.exists(path):
            print(f'Failed to locate "{path}"!', file=sys.stderr)
            sys.exit(1)
        if os.path.isdir(path):
            # Exploring existing isolated tracks
            if len([f for f in os.listdir(path) if is_isolated_track(f)]) < 2:
                print(f'Not enough isolated tracks to explore in "{path}"!', file=sys.stderr)
                sys.exit(1)
            input_dir_path = path
        else:
            # Creating new isolated tracks from a song (or no song provided, which will prompt the user to choose)
            song_path = path
    
    print('Launching Unmixer UI...')
    UnmixerUI(song_path=song_path, input_dir_path=input_dir_path).run()


if __name__ == '__main__':
    main()



APP_NAME = 'Unmixer'
ORGANIZATION_NAME = 'DyeSoft'
PROJECT_README_URL = 'https://github.com/will2dye4/unmixer#usage'

ERROR_MESSAGE_TITLE = 'Error'
SUCCESS_MESSAGE_TITLE = 'Success'

OK_BUTTON_TEXT = 'OK'

OTHER_TRACK_NAME_DISABLED_TOOLTIP = 'Select a different pre-trained model to change this setting.'

FONT_WEIGHT_BOLD = 800



from typing import Optional
import os.path
import shutil

from PyQt6.QtCore import (
    pyqtSignal,
    QLine,
    QObject,
    QPoint,
    QPointF,
    QRect,
    Qt,
    QThread,
    QTimer,
    QUrl,
)
from PyQt6.QtGui import QMouseEvent, QPainter, QPainterPath
from PyQt6.QtMultimedia import QAudioOutput, QMediaPlayer
from PyQt6.QtWidgets import (
    QApplication,
    QFileDialog,
    QHBoxLayout,
    QLabel,
    QMessageBox,
    QPushButton,
    QSlider,
    QVBoxLayout,
    QWidget,
)
from soundfile import SoundFile

from unmixer.constants import MAX_VOLUME, MIN_VOLUME, settings
from unmixer.ui.constants import (
    ERROR_MESSAGE_TITLE,
    FONT_WEIGHT_BOLD,
    SUCCESS_MESSAGE_TITLE,
)
from unmixer.ui.track import Track
from unmixer.ui.waveform import Waveform, WAVEFORM_BACKGROUND_COLORS


MINUTES_PER_HOUR = 60
SECONDS_PER_MINUTE = MINUTES_PER_HOUR
SECONDS_PER_HOUR = SECONDS_PER_MINUTE * MINUTES_PER_HOUR


class PlaybackControls(QWidget):

    PLAYBACK_CONTROL_FONT_SIZE = 30
    PLAYBACK_CONTROL_BUTTON_SPACING = 0
    PLAYBACK_CONTROL_BUTTON_STYLE = 'background: transparent; border: none'

    PLAYBACK_TIME_FONT_SIZE = 15
    DEFAULT_PLAYBACK_TIME_TEXT = '0:00 / 0:00'
    PLAYBACK_TIMER_INTERVAL_MILLIS = 500
    
    LAYOUT_SPACING = 50

    MIN_BUTTON_WIDTH = 90
    MIN_PLAYBACK_TIME_WIDTH = 120

    VOLUME_INCREMENT = 5

    SKIP_INTERVAL_MILLIS = 1_000  # Skip forward/back one second at a time.
    SKIP_TIMER_INTERVAL_MILLIS = 200  # Skip forward/back every 200 ms when the corresponding button is held down.

    EXPORT_BUTTON_TEXT = '💾 Export...'
    PLAY_BUTTON_TEXT = '▶️'
    PAUSE_BUTTON_TEXT = '⏸️'
    RESTART_BUTTON_TEXT = '⏮️'
    SKIP_BACK_BUTTON_TEXT = '⏪️'
    SKIP_FORWARD_BUTTON_TEXT = '⏩️'
    
    EXPORT_BUTTON_ENABLED_TOOLTIP = 'Export the selected tracks to a new file.'
    EXPORT_BUTTON_DISABLED_TOOLTIP_TEMPLATE = 'Enable 2-{n} tracks to export them as a new mix.'
    EXPORT_BUTTON_NOT_ENOUGH_TRACKS_TOOLTIP = 'There are not enough tracks to create a new mix.'

    def __init__(self, parent: QWidget) -> None:
        super().__init__(parent)
        
        self.export_button = QPushButton(self.EXPORT_BUTTON_TEXT)
        self.export_button.setDisabled(True)  # Initially, all tracks are selected, so exporting is disabled.
        if len(self.parent().tracks) < 3:
            self.export_button_disabled_tooltip = self.EXPORT_BUTTON_NOT_ENOUGH_TRACKS_TOOLTIP
        else:
            self.export_button_disabled_tooltip = self.EXPORT_BUTTON_DISABLED_TOOLTIP_TEMPLATE.format(n=len(parent.tracks)-1)
        self.export_button.setStatusTip(self.export_button_disabled_tooltip)
        self.export_button.setToolTip(self.export_button_disabled_tooltip)
        self.export_button.setMinimumWidth(self.MIN_BUTTON_WIDTH)
        self.export_button.clicked.connect(self.export_selected_tracks)

        self.restart_button = QPushButton(self.RESTART_BUTTON_TEXT)
        self.restart_button.setStyleSheet(self.PLAYBACK_CONTROL_BUTTON_STYLE)
        font = self.restart_button.font()
        font.setPixelSize(self.PLAYBACK_CONTROL_FONT_SIZE)
        self.restart_button.setFont(font)
        self.restart_button.clicked.connect(self.restart_track)

        self._skipping_back = False
        self.skip_back_button = QPushButton(self.SKIP_BACK_BUTTON_TEXT)
        self.skip_back_button.setStyleSheet(self.PLAYBACK_CONTROL_BUTTON_STYLE)
        font = self.skip_back_button.font()
        font.setPixelSize(self.PLAYBACK_CONTROL_FONT_SIZE)
        self.skip_back_button.setFont(font)
        self.skip_back_button.mousePressEvent = self.skip_back_button_pressed
        self.skip_back_button.mouseReleaseEvent = self.skip_back_button_released
        self.skip_back_timer = QTimer(self)
        self.skip_back_timer.setInterval(self.SKIP_TIMER_INTERVAL_MILLIS)
        self.skip_back_timer.timeout.connect(self.skip_back)

        self.playing = False
        self.play_button = QPushButton(self.PLAY_BUTTON_TEXT)
        self.play_button.setStyleSheet(self.PLAYBACK_CONTROL_BUTTON_STYLE)
        font = self.play_button.font()
        font.setPixelSize(int(self.PLAYBACK_CONTROL_FONT_SIZE * 1.5))  # Make Play/Pause button bigger than the rest.
        self.play_button.setFont(font)
        self.play_button.clicked.connect(self.toggle_playback)

        self._skipping_forward = False
        self.skip_forward_button = QPushButton(self.SKIP_FORWARD_BUTTON_TEXT)
        self.skip_forward_button.setStyleSheet(self.PLAYBACK_CONTROL_BUTTON_STYLE)
        font = self.skip_forward_button.font()
        font.setPixelSize(self.PLAYBACK_CONTROL_FONT_SIZE)
        self.skip_forward_button.setFont(font)
        self.skip_forward_button.mousePressEvent = self.skip_forward_button_pressed
        self.skip_forward_button.mouseReleaseEvent = self.skip_forward_button_released
        self.skip_forward_timer = QTimer(self)
        self.skip_forward_timer.setInterval(self.SKIP_TIMER_INTERVAL_MILLIS)
        self.skip_forward_timer.timeout.connect(self.skip_forward)

        self.volume_slider = QSlider(Qt.Orientation.Horizontal)
        self.volume_slider.setMinimum(MIN_VOLUME)
        self.volume_slider.setMaximum(MAX_VOLUME)
        self.volume_slider.setValue(self.parent()._volume)  # parent.volume property does not exist yet
        self.volume_slider.valueChanged.connect(self.volume_changed)

        self.playback_time = QLabel(self.DEFAULT_PLAYBACK_TIME_TEXT)
        font = self.playback_time.font()
        font.setPixelSize(self.PLAYBACK_TIME_FONT_SIZE)
        font.setWeight(FONT_WEIGHT_BOLD)
        self.playback_time.setFont(font)
        self.playback_time.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.playback_time.setMinimumWidth(self.MIN_PLAYBACK_TIME_WIDTH)
        self.update_playback_time_timer = QTimer(self)
        self.update_playback_time_timer.setInterval(self.PLAYBACK_TIMER_INTERVAL_MILLIS)
        self.update_playback_time_timer.timeout.connect(self.update_playback_time)
        self.update_playback_time_timer.start()

        layout = QHBoxLayout()
        layout.addStretch()
        layout.addWidget(self.export_button)
        layout.addSpacing(self.LAYOUT_SPACING)

        sublayout = QHBoxLayout()
        sublayout.setSpacing(self.PLAYBACK_CONTROL_BUTTON_SPACING)
        sublayout.addWidget(self.restart_button)
        sublayout.addWidget(self.skip_back_button)
        sublayout.addWidget(self.play_button)
        sublayout.addWidget(self.skip_forward_button)
        layout.addLayout(sublayout)

        layout.addWidget(self.volume_slider)
        layout.addSpacing(self.LAYOUT_SPACING)
        layout.addWidget(self.playback_time)
        layout.addStretch()

        self.setLayout(layout)
        self.show()

    @property
    def player(self) -> QMediaPlayer:
        return self.parent().player

    @staticmethod
    def format_playback_time(time_in_millis: int) -> str:
        total_seconds = round(time_in_millis / 1000)
        hours, remaining_seconds = divmod(total_seconds, SECONDS_PER_HOUR)
        minutes, seconds = divmod(remaining_seconds, SECONDS_PER_MINUTE)
        return f'{hours}:{minutes:02}:{seconds:02}' if hours else f'{minutes}:{seconds:02}'

    def update_playback_time(self) -> None:
        if self.player and self.player.isPlaying():
            current_time = self.format_playback_time(self.player.position())
            total_time = self.format_playback_time(self.player.duration())
            self.playback_time.setText(f'{current_time} / {total_time}')
            self.update()

    def toggle_playback_state(self) -> None:
        self.playing = not self.playing
        self.play_button.setText(self.PAUSE_BUTTON_TEXT if self.playing else self.PLAY_BUTTON_TEXT)
        self.update()

    def toggle_playback(self) -> None:
        if not self.play_button.isEnabled():
            return

        self.toggle_playback_state()
        if self.playing:
            self.parent().play_selected_tracks()
        else:
            self.parent().pause_playback()
        self.update()

    def restart_track(self) -> None:
        if self.player:
            if self.player.isPlaying():
                self.player.setPosition(0)
            else:
                self.parent().prev_position = 0
        self.playback_time.setText(self.DEFAULT_PLAYBACK_TIME_TEXT)
        self.update()
    
    def skip_back_button_pressed(self, event: QMouseEvent) -> None:
        if self.player and self.player.isPlaying() and not self._skipping_back and not self._skipping_forward:
            self._skipping_back = True
            # Skip back once in case the button is pressed and released quickly (i.e., a normal click).
            self.skip_back()
            self.skip_back_timer.start()
        event.accept()

    def skip_back_button_released(self, event: QMouseEvent) -> None:
        if self._skipping_back:
            self._skipping_back = False
            self.skip_back_timer.stop()
        event.accept()

    def skip_forward_button_pressed(self, event: QMouseEvent) -> None:
        if self.player and self.player.isPlaying() and not self._skipping_back and not self._skipping_forward:
            self._skipping_forward = True
            # Skip forward once in case the button is pressed and released quickly (i.e., a normal click).
            self.skip_forward()
            self.skip_forward_timer.start()
        event.accept()

    def skip_forward_button_released(self, event: QMouseEvent) -> None:
        if self._skipping_forward:
            self._skipping_forward = False
            self.skip_forward_timer.stop()
        event.accept()

    def export_selected_tracks(self) -> None:
        self.parent().export_selected_tracks()

    def enable_export_button(self) -> None:
        self.export_button.setDisabled(False)
        self.export_button.setStatusTip(self.EXPORT_BUTTON_ENABLED_TOOLTIP)
        self.export_button.setToolTip(self.EXPORT_BUTTON_ENABLED_TOOLTIP)
        self.parent().parent().enable_export_menu_action()

    def disable_export_button(self) -> None:
        self.export_button.setDisabled(True)
        self.export_button.setStatusTip(self.export_button_disabled_tooltip)
        self.export_button.setToolTip(self.export_button_disabled_tooltip)
        self.parent().parent().disable_export_menu_action()
    
    def enable_play_button(self) -> None:
        self.play_button.setDisabled(False)

    def disable_play_button(self) -> None:
        self.play_button.setDisabled(True)

    def volume_changed(self) -> None:
        self.parent().volume = max(MIN_VOLUME, min(self.volume_slider.value(), MAX_VOLUME))

    def increase_volume(self) -> None:
        if (volume := self.parent().volume) < MAX_VOLUME:
            self.volume_slider.setValue(min(volume + self.VOLUME_INCREMENT, MAX_VOLUME))

    def decrease_volume(self) -> None:
        if (volume := self.parent().volume) > MIN_VOLUME:
            self.volume_slider.setValue(max(MIN_VOLUME, volume - self.VOLUME_INCREMENT))

    def skip_forward(self) -> None:
        if self.player and self.player.isPlaying() and (position := self.player.position()) < self.player.duration():
            self.player.setPosition(min(position + self.SKIP_INTERVAL_MILLIS, self.player.duration()))

    def skip_back(self) -> None:
        if self.player and self.player.isPlaying() and (position := self.player.position()) > 0:
            self.player.setPosition(max(0, position - self.SKIP_INTERVAL_MILLIS))


class MultiTrackPlayhead(QWidget):
    
    COLOR = Waveform.FOREGROUND_COLOR
    DRAG_SCALE_FACTOR = 0.9
    PLAYHEAD_SIZE = 16
    TIMER_INTERVAL_MILLIS = 500
    
    def __init__(self, parent: QWidget) -> None:
        super().__init__(parent)
        self.setAttribute(Qt.WidgetAttribute.WA_Hover)
        
        self._dragging = False
        
        self.update_playhead_timer = QTimer(self)
        self.update_playhead_timer.setInterval(self.TIMER_INTERVAL_MILLIS)
        self.update_playhead_timer.timeout.connect(self.repaint)
        self.update_playhead_timer.start()

        self.raise_()  # Bring the playhead to the top of the "stack" (z-axis).
        self.show()
        
    @property
    def player(self) -> QMediaPlayer:
        return self.parent().player
    
    @property
    def progress(self) -> float:
        if self.player:
            return self.player.position() / self.player.duration()
        return 0
    
    def adjust_size_and_position(self) -> None:
        x1, y1, x2, y2 = self.parent().geometry().getCoords()
        track_controls_width = self.parent().tracks[0].controls.width()
        play_controls_height = self.parent().controls.height()
        self.setGeometry(QRect(x1 - track_controls_width, y1, x2, y2 - play_controls_height))
        self.move(x1 + track_controls_width, y1)
    
    def is_mouse_event_within_playhead(self, event) -> bool:
        top_waveform = self.parent().tracks[0].waveform
        line_top = self.mapFromGlobal(top_waveform.mapToGlobal(QPoint(int(top_waveform.width() * self.progress), 0)))
        playhead_bounds = QRect(int(line_top.x() - self.PLAYHEAD_SIZE / 2), line_top.y() - self.PLAYHEAD_SIZE,
                                int(line_top.x() + self.PLAYHEAD_SIZE / 2), line_top.y())
        return playhead_bounds.contains(event.pos())
    
    def mousePressEvent(self, event):
        if self.is_mouse_event_within_playhead(event):
            self._dragging = True
            QApplication.setOverrideCursor(Qt.CursorShape.ClosedHandCursor)
            QApplication.processEvents()
    
    def mouseMoveEvent(self, event):
        if not self._dragging or not self.player:
            return
        
        top_waveform = self.parent().tracks[0].waveform
        position = int((event.pos().x() / top_waveform.width()) * self.player.duration() * self.DRAG_SCALE_FACTOR)
        if self.player.isPlaying():
            self.player.setPosition(position)
        else:
            self.parent().prev_position = position
    
    def mouseReleaseEvent(self, event):
        if self._dragging:
            self._dragging = False
            QApplication.restoreOverrideCursor()
            QApplication.processEvents()
    
    def paintEvent(self, event) -> None:
        self.adjust_size_and_position()
        painter = QPainter()
        painter.begin(self)
        self.draw_playhead(painter)
        painter.end()
    
    def draw_playhead(self, painter: QPainter) -> None:
        painter.setRenderHint(QPainter.RenderHint.Antialiasing)
        pen = painter.pen()
        pen.setColor(self.COLOR)
        pen.setWidth(2)
        painter.setPen(pen)

        # Draw vertical line.
        top_waveform = self.parent().tracks[0].waveform
        line_top = self.mapFromGlobal(top_waveform.mapToGlobal(QPoint(int(top_waveform.width() * self.progress), 0)))
        bottom_waveform = self.parent().tracks[-1].waveform
        line_bottom = self.mapFromGlobal(bottom_waveform.mapToGlobal(QPoint(int(top_waveform.width() * self.progress),
                                                                            bottom_waveform.height())))
        painter.drawLine(QLine(line_top.x(), line_top.y(), line_bottom.x(), line_bottom.y()))
        
        # Draw triangular playhead at the top.
        triangle_path = QPainterPath()
        triangle_path.moveTo(line_top.toPointF())
        triangle_path.lineTo(QPointF(line_top.x() - self.PLAYHEAD_SIZE / 2, line_top.y() - self.PLAYHEAD_SIZE))
        triangle_path.lineTo(QPointF(line_top.x() + self.PLAYHEAD_SIZE / 2, line_top.y() - self.PLAYHEAD_SIZE))
        triangle_path.lineTo(line_top.toPointF())
        painter.fillPath(triangle_path, self.COLOR)


# Reference: https://realpython.com/python-pyqt-qthread/#using-qthread-vs-pythons-threading
class UpdateMediaPlayerWorker(QObject):
    exception = pyqtSignal(Exception)
    finished = pyqtSignal()

    def __init__(self, display: 'MultiTrackDisplay') -> None:
        super().__init__()
        self.display = display

    def run(self):
        try:
            # This will create the file using ffmpeg, assuming it doesn't exist already.
            self.display.file_path_for_selected_tracks = self.display.parent().temp_file_path(
                self.display.selected_track_paths
            )
        except Exception as e:
            self.exception.emit(e)
        else:
            self.finished.emit()


class MultiTrackDisplay(QWidget):
    
    EXPORT_DIALOG_TITLE = 'Export Selected Tracks'
    
    TITLE_FONT_SIZE = 30
    
    def __init__(self, parent: QWidget, song_title: str, file_paths: list[str]) -> None:
        super().__init__(parent)
        self.file_path_for_selected_tracks = None
        self._ffmpeg_thread = None
        self._ffmpeg_thread_worker = None
        self._soloed_track = None
        self._volume = self.parent().app.setting(settings.playback.VOLUME)
        
        self.title = QLabel(f'Song:  {song_title}')
        font = self.title.font()
        font.setPixelSize(self.TITLE_FONT_SIZE)
        font.setWeight(FONT_WEIGHT_BOLD)
        self.title.setFont(font)
        self.title.setAlignment(Qt.AlignmentFlag.AlignCenter)
        
        self.audio_output = None
        self.player = None
        self.prev_position = None
        
        self.tracks = []
        for i, file_path in enumerate(file_paths):
            with SoundFile(file_path, 'rb') as sound_file:
                colors = WAVEFORM_BACKGROUND_COLORS[i % len(WAVEFORM_BACKGROUND_COLORS)]
                self.tracks.append(Track(sound_file, colors))
                
        self.controls = PlaybackControls(self)  # This needs to be initialized AFTER self.tracks is populated!
        
        layout = QVBoxLayout()
        layout.addWidget(self.title, stretch=1)
        for track in self.tracks:
            layout.addWidget(track, stretch=3)
        layout.addWidget(self.controls, stretch=1)
        self.setLayout(layout)
        
        self.playhead = MultiTrackPlayhead(self)  # This needs to be initialized AFTER self.tracks is populated!
        
        self.setFocusProxy(self.controls.play_button)
        self.show()
    
    @property
    def selected_tracks(self) -> list[Track]:
        if self.soloed_track:
            return [self.soloed_track]
        return [track for track in self.tracks if not track.muted]
    
    @property
    def selected_track_paths(self) -> list[str]:
        return [track.file_path for track in self.selected_tracks]

    @property
    def soloed_track(self) -> Optional[Track]:
        return self._soloed_track
    
    @soloed_track.setter
    def soloed_track(self, track: Optional[Track]) -> None:
        if track:
            track.enable_waveform()
            for other_track in self.tracks:
                if other_track != track:
                    if other_track == self.soloed_track:
                        other_track.controls.toggle_soloed_state()
                    other_track.disable_waveform()
        else:
            for other_track in self.tracks:
                if not other_track.muted:
                    other_track.enable_waveform()
        self._soloed_track = track
    
    @property
    def volume(self) -> int:
        return self._volume

    @volume.setter
    def volume(self, value: int) -> None:
        self._volume = value
        self.parent().app.update_setting(settings.playback.VOLUME, value)
        if self.player:
            self.audio_output.setVolume(self._volume / MAX_VOLUME)
            self.player.setAudioOutput(self.audio_output)
    
    def handle_media_status_changed(self, status: QMediaPlayer.MediaStatus) -> None:
        if status == QMediaPlayer.MediaStatus.EndOfMedia and self.controls.playing:
            # Auto-pause at the end of the track.
            self.controls.toggle_playback_state()
    
    def init_audio(self, source_url: Optional[QUrl] = None) -> tuple[QAudioOutput, QMediaPlayer]:
        audio_output = QAudioOutput()
        audio_output.setVolume(self._volume / MAX_VOLUME)
        player = QMediaPlayer()
        player.mediaStatusChanged.connect(self.handle_media_status_changed)
        player.setAudioOutput(audio_output)
        if source_url:
            player.setSource(source_url)
        return audio_output, player
    
    def update_media_player_for_current_selected_tracks(self) -> None:
        selected_tracks = [track.file_path for track in self.selected_tracks]
        if self.parent().has_temp_mix_file(selected_tracks):
            # If the mix file already exists, it's fine to get the file path synchronously.
            self.file_path_for_selected_tracks = self.parent().temp_file_path(selected_tracks)
            self.update_media_player_for_current_selected_tracks_sync()
            return

        # If the mix file does not already exist, it will first be created using ffmpeg.
        # Show a status message, then spawn a thread to handle the creation of the file
        # so that UI updates are unblocked.
        filename = os.path.basename(self.parent().output_file_path(selected_tracks))
        self.parent().show_status_message(f'Mixing {filename}...')

        self._ffmpeg_thread = QThread()
        self._ffmpeg_thread_worker = UpdateMediaPlayerWorker(self)
        self._ffmpeg_thread.started.connect(self._ffmpeg_thread_worker.run)
        self._ffmpeg_thread.finished.connect(self._ffmpeg_thread.deleteLater)
        self._ffmpeg_thread.finished.connect(self.update_media_player_for_current_selected_tracks_sync)

        self._ffmpeg_thread_worker.moveToThread(self._ffmpeg_thread)
        self._ffmpeg_thread_worker.finished.connect(self._ffmpeg_thread.quit)
        self._ffmpeg_thread_worker.finished.connect(self._ffmpeg_thread_worker.deleteLater)
        self._ffmpeg_thread_worker.exception.connect(self.handle_ffmpeg_thread_exception)

        self._ffmpeg_thread.start()

    def update_media_player_for_current_selected_tracks_sync(self) -> None:
        if not self.selected_tracks or not self.controls.playing or not self.file_path_for_selected_tracks:
            return
        
        source_url = QUrl.fromLocalFile(self.file_path_for_selected_tracks)
        if not self.player or self.player.source() != source_url:
            audio_output, player = self.init_audio(source_url)
            if self.player and self.player.isPlaying():
                self.prev_position = self.player.position()
                QTimer.singleShot(0, self.player.stop)  # https://stackoverflow.com/a/49239010
            self.audio_output = audio_output
            self.player = player
        
        if self.controls.playing and not self.player.isPlaying():
            if self.prev_position is not None:
                # Seek the new track to the same position as the previous track.
                self.player.setPosition(self.prev_position)
            self.player.play()
    
    def play_selected_tracks(self) -> None:
        if self.player and self.player.isPlaying():
            return
        self.update_media_player_for_current_selected_tracks()
        self.prev_position = None
    
    def pause_playback(self) -> None:
        if self.player and self.player.isPlaying():
            self.prev_position = self.player.position()
            self.player.pause()
    
    def export_selected_tracks(self) -> None:
        # If a single track, no tracks, or all tracks are selected, then there is nothing to export.
        if len(self.selected_tracks) < 2 or len(self.selected_tracks) == len(self.tracks):
            return
        
        temp_file_path = self.parent().temp_file_path(self.selected_track_paths)
        _, extension = os.path.splitext(temp_file_path)
        export_dir = os.path.dirname(self.selected_tracks[0].file_path)
        
        names = sorted(os.path.splitext(os.path.basename(track.file_path))[0] for track in self.selected_tracks)
        export_name = '+'.join(names) + extension
        default_export_path = os.path.join(export_dir, export_name)
        export_path, _ = QFileDialog.getSaveFileName(self, self.EXPORT_DIALOG_TITLE, default_export_path)
        if not export_path:
            return  # User canceled or did not enter a path.
        
        if not export_path.endswith(extension):
            # Force file path to end with the correct extension.
            export_path = os.path.splitext(export_path)[0] + extension
        
        # This will overwrite the file if it exists, but the user already confirmed that they want that.
        shutil.copy(temp_file_path, export_path)
        QMessageBox.information(self, SUCCESS_MESSAGE_TITLE, f'Successfully exported {export_path}.')
        print(f'Exported new mix file "{export_path}".')

    def handle_ffmpeg_thread_exception(self, exception: Exception) -> None:
        print(f'Failed to create temporary mix file: {exception}')
        QMessageBox.warning(self, ERROR_MESSAGE_TITLE, 'Failed to create mix!')



from typing import Any, Callable, Optional
import abc
import os.path
import tempfile

from PyQt6.QtCore import QSettings, QSize, Qt, QUrl
from PyQt6.QtGui import (
    QAction,
    QCloseEvent,
    QDesktopServices,
    QDragEnterEvent,
    QDragMoveEvent,
    QDropEvent,
    QKeySequence,
)
from PyQt6.QtWidgets import (
    QApplication,
    QFileDialog,
    QLabel,
    QMainWindow,
    QMenu,
    QMenuBar,
    QMessageBox,
    QScrollArea,
)

from unmixer.constants import (
    DEFAULT_OTHER_TRACK_NAME,
    DEFAULT_SETTINGS,
    models,
    settings,
)
from unmixer.remix import merge_audio_files
from unmixer.ui.constants import (
    APP_NAME,
    ERROR_MESSAGE_TITLE,
    ORGANIZATION_NAME,
    PROJECT_README_URL,
    SUCCESS_MESSAGE_TITLE,
)
from unmixer.ui.importer import SongImporter
from unmixer.ui.multitrack import MultiTrackDisplay
from unmixer.ui.preferences import UnmixerPreferences
from unmixer.ui.track import Track
from unmixer.util import cleanup_intermediate_dir, expand_path, is_isolated_track


class UnmixerMainWindow(QMainWindow):

    EXPORT_MENU_ACTION_KEY = 'export'
    IMPORT_MENU_ACTION_KEY = 'import'
    PLAY_MENU_ACTION_KEY = 'play'

    STATUS_BAR_FONT_SIZE = 10
    STATUS_MESSAGE_DURATION_MILLIS = 1_000

    def __init__(self, app: 'UnmixerUI', show_status_bar: bool = True) -> None:
        super().__init__()
        self.app = app
        self.menu_bar = QMenuBar(self)
        self._menu_actions = {}

        if show_status_bar:
            status_bar_font = self.statusBar().font()
            status_bar_font.setPixelSize(self.STATUS_BAR_FONT_SIZE)
            self.statusBar().setFont(status_bar_font)
            self.statusBar().setMaximumHeight(int(self.STATUS_BAR_FONT_SIZE * 1.5))

    @abc.abstractmethod
    def create_menu_bar(self) -> None:
        raise NotImplemented

    def focusInEvent(self, event):
        self.app.active_window = self

    def closeEvent(self, event: QCloseEvent) -> None:
        try:
            for window in list(self.app.track_windows.values()):
                if window is self:
                    if window.input_dir_path in self.app.track_windows:
                        del self.app.track_windows[window.input_dir_path]
                else:
                    window.reset_menu_bar()
            if self.app.import_window:
                if self.app.import_window is self:
                    self.app.import_window = None
                else:
                    self.app.import_window.reset_menu_bar()
            if self.app.preferences_window:
                if self.app.preferences_window is self:
                    self.app.preferences_window = None
                else:
                    self.app.preferences_window.reset_menu_bar()
        finally:
            event.accept()

    def reset_menu_bar(self) -> None:
        self.menu_bar.clear()
        self.create_menu_bar()

    def create_file_menu(self) -> QMenu:
        file_menu = self.menu_bar.addMenu('&File')
        self.menu_bar.addAction(file_menu.menuAction())

        # File > Open Isolated Tracks...
        open_tracks_action = file_menu.addAction('&Open Isolated Tracks...')
        open_tracks_action.setShortcut(QKeySequence.StandardKey.Open)
        open_tracks_action.triggered.connect(self.app.choose_track_directory)

        # File > Open Recent
        recent_menu = file_menu.addMenu('Open Recent')
        file_menu.addAction(recent_menu.menuAction())

        # File > Open Recent > [Song Title]
        recently_opened = self.app.settings.value(settings.open.RECENTLY_OPENED, [], 'QStringList')
        removed = False
        for recent_dir_path in recently_opened:
            if not os.path.exists(recent_dir_path):
                recently_opened.remove(recent_dir_path)
                removed = True
                continue
            song_title = os.path.basename(recent_dir_path.rstrip(os.path.sep))
            open_action = recent_menu.addAction(song_title)
            open_action.setStatusTip(recent_dir_path)
            open_action.setToolTip(recent_dir_path)
            open_action.triggered.connect(self.open_recent_menu_action(recent_dir_path))
        
        if removed:
            self.app.update_setting(settings.open.RECENTLY_OPENED, recently_opened)

        return file_menu

    def add_close_action_to_file_menu(self, file_menu: QMenu) -> None:
        # File > Close
        file_menu.addSeparator()
        close_window_action = file_menu.addAction('&Close')
        close_window_action.setShortcut(QKeySequence.StandardKey.Close)
        close_window_action.triggered.connect(self.close)

    def create_window_menu(self) -> None:
        window_menu = self.menu_bar.addMenu('&Window')
        self.menu_bar.addAction(window_menu.menuAction())

        # Window > Song Importer
        song_importer_action = window_menu.addAction('&Song Importer')
        song_importer_action.setShortcut(QKeySequence('Ctrl+0'))
        song_importer_action.triggered.connect(self.app.bring_import_window_to_front)

        # Window > Track Explorer - [Song Title]
        if self.app.track_windows or isinstance(self, UnmixerTrackExplorerWindow):
            window_menu.addSeparator()
            track_windows = list(self.app.track_windows.values())
            if self not in track_windows and isinstance(self, UnmixerTrackExplorerWindow):
                track_windows.append(self)
            for i, track_window in enumerate(sorted(track_windows, key=lambda w: w.song_title)):
                track_explorer_action = window_menu.addAction(f'&Track Explorer - {track_window.song_title}')
                if (index := i + 1) < 10:
                    track_explorer_action.setShortcut(QKeySequence(f'Ctrl+T,Ctrl+{index}'))
                track_explorer_action.triggered.connect(track_window.bring_to_front)

    def create_help_menu(self) -> None:
        help_menu = self.menu_bar.addMenu('&Help')
        self.menu_bar.addAction(help_menu.menuAction())

        # Help > About...
        # NOTE: On macOS, this will appear under the application (Python) menu instead of the Help menu.
        about_action = help_menu.addAction('&About...')
        about_action.setMenuRole(QAction.MenuRole.AboutRole)
        about_action.setShortcut(QKeySequence.StandardKey.HelpContents)
        about_action.triggered.connect(self.app.open_project_readme)

        # Help > Preferences...
        # NOTE: On macOS, this will appear under the application (Python) menu instead of the Help menu.
        preferences_action = help_menu.addAction('&Preferences...')
        preferences_action.setMenuRole(QAction.MenuRole.PreferencesRole)
        preferences_action.setShortcut(QKeySequence.StandardKey.Preferences)
        preferences_action.triggered.connect(self.app.bring_preferences_window_to_front)

    def open_recent_menu_action(self, dir_path: str) -> Callable[[], None]:
        def handler():
            self.app.show_track_explorer_window(dir_path)
        return handler

    def bring_to_front(self) -> None:
        self.raise_()
        self.activateWindow()
        self.setFocus()

    def show_status_message(self, message: str) -> None:
        self.statusBar().showMessage(message, self.STATUS_MESSAGE_DURATION_MILLIS)
        self.update()
        QApplication.processEvents()


class UnmixerPreferencesWindow(UnmixerMainWindow):

    WINDOW_TITLE = f'{APP_NAME} | Preferences'

    MIN_HEIGHT = 600
    MIN_WIDTH = MIN_HEIGHT

    DEFAULT_HEIGHT = 850
    DEFAULT_WIDTH = 700

    def __init__(self, app: 'UnmixerUI') -> None:
        super().__init__(app)
        self.preferences = UnmixerPreferences(self, app)
        self.scroll_area = QScrollArea()
        self.scroll_area.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAsNeeded)
        self.scroll_area.setWidgetResizable(True)
        self.scroll_area.setWidget(self.preferences)
        self.setCentralWidget(self.scroll_area)
        self.setWindowTitle(self.WINDOW_TITLE)
        self.setMinimumSize(self.MIN_WIDTH, self.MIN_HEIGHT)
        self.resize(QSize(self.DEFAULT_WIDTH, self.DEFAULT_HEIGHT))
        self.create_menu_bar()

    def create_menu_bar(self) -> None:
        # Window menu
        self.create_window_menu()

        # Help menu
        self.create_help_menu()

    def set_audio_format(self, audio_format: str) -> None:
        self.preferences.set_audio_format(audio_format)

    def set_other_track_name(self, other_track_name: str) -> None:
        self.preferences.set_other_track_name(other_track_name)


class UnmixerImportWindow(UnmixerMainWindow):

    DEFAULT_WINDOW_TITLE = f'{APP_NAME} | Select a Song'
    IMPORTING_WINDOW_TITLE = f'{APP_NAME} | Importing...'

    MIN_HEIGHT = 400
    MIN_WIDTH = 600

    def __init__(self, app: 'UnmixerUI', source_file_path: Optional[str] = None) -> None:
        super().__init__(app, show_status_bar=False)
        self.importer = SongImporter(app, source_file_path)
        self.setCentralWidget(self.importer)
        self.setWindowTitle(self.DEFAULT_WINDOW_TITLE)
        self.setMinimumSize(self.MIN_WIDTH, self.MIN_HEIGHT)
        self.setAcceptDrops(True)
        self.create_menu_bar()

    def dragEnterEvent(self, event: QDragEnterEvent) -> None:
        if self.importer.import_process is not None and self.importer.import_process.is_alive():
            event.ignore()
        else:
            event.acceptProposedAction()

    def dragMoveEvent(self, event: QDragMoveEvent) -> None:
        if self.importer.import_process is not None and self.importer.import_process.is_alive():
            event.ignore()
        else:
            event.acceptProposedAction()

    def dropEvent(self, event: QDropEvent) -> None:
        input_path = None
        for url in event.mimeData().urls():
            if url.isLocalFile():
                input_path = url.toLocalFile()
                event.acceptProposedAction()
                break

        if input_path is None:
            event.ignore()
            return

        if os.path.isdir(input_path):
            self.app.output_dir_path = input_path
            self.app.update_setting(settings.open.TRACK_DIR_PATH, os.path.dirname(input_path))
            self.app.show_track_explorer_window(input_path)
        else:
            self.app.song_path = input_path
            self.importer.input_file_path = input_path
            self.app.update_setting(settings.importer.DIR_PATH, os.path.dirname(input_path))
            self.importer.select_song()

    def create_menu_bar(self) -> None:
        # File menu
        file_menu = self.create_file_menu()
        file_menu.addSeparator()

        # File > Import Song...
        import_song_action = file_menu.addAction('&Import Song...')
        import_song_action.setShortcut(QKeySequence('Ctrl+I'))
        import_song_action.triggered.connect(self.importer.choose_file)
        self._menu_actions[self.IMPORT_MENU_ACTION_KEY] = import_song_action

        # File > Close
        self.add_close_action_to_file_menu(file_menu)

        # Window menu
        self.create_window_menu()

        # Help menu
        self.create_help_menu()

    def enable_import_menu_action(self) -> None:
        self._menu_actions[self.IMPORT_MENU_ACTION_KEY].setDisabled(False)

    def disable_import_menu_action(self) -> None:
        self._menu_actions[self.IMPORT_MENU_ACTION_KEY].setDisabled(True)

    def set_audio_format(self, audio_format: str) -> None:
        self.importer.import_settings.set_audio_format(audio_format)

    def set_other_track_name(self, other_track_name: str) -> None:
        self.importer.import_settings.set_other_track_name(other_track_name)

    def reset(self) -> None:
        self.setWindowTitle(self.DEFAULT_WINDOW_TITLE)
        self.reset_menu_bar()
        self.importer.reset()
        self.update()


class UnmixerTrackExplorerWindow(UnmixerMainWindow):

    WINDOW_TITLE = f'{APP_NAME} | Track Explorer'

    MIN_HEIGHT = 800
    MIN_WIDTH = 1000

    def __init__(self, app: 'UnmixerUI', input_dir_path: str, source_file_path: Optional[str] = None) -> None:
        super().__init__(app)

        # Optimization only: store the source file path after importing a file
        # to avoid the need to create a full mix on-the-fly using ffmpeg.
        self.source_file_path = expand_path(source_file_path) if source_file_path else None
        self.input_dir_path = expand_path(input_dir_path)
        self.song_title = os.path.basename(self.input_dir_path.rstrip(os.path.sep))
        file_paths = sorted(
            os.path.join(self.input_dir_path, filename)
            for filename in os.listdir(self.input_dir_path)
            if is_isolated_track(filename)
        )
        if len(file_paths) < 2:
            raise ValueError(f'Not enough isolated tracks to explore in "{self.input_dir_path}"!')

        self.tracks = MultiTrackDisplay(self, self.song_title, file_paths)
        self.setCentralWidget(self.tracks)
        self.setWindowTitle(f'{self.WINDOW_TITLE} - {self.song_title}')
        self.setMinimumSize(self.MIN_WIDTH, self.MIN_HEIGHT)
        self.create_menu_bar()
        
        self.statusBar().addPermanentWidget(QLabel(self.input_dir_path))

        self._temp_dir = None
        self._temp_file_paths = set()

    @property
    def temp_dir(self) -> tempfile.TemporaryDirectory:
        if self._temp_dir is None:
            self._temp_dir = tempfile.TemporaryDirectory(prefix=f'unmixer_{self.song_title.replace(" ", "_")}_')
        return self._temp_dir

    @property
    def temp_dir_path(self) -> str:
        return self.temp_dir.name

    def closeEvent(self, event: QCloseEvent):
        if self.tracks.player and self.tracks.player.isPlaying():
            self.tracks.player.stop()
        super().closeEvent(event)

    def create_menu_bar(self) -> None:
        # File menu
        file_menu = self.create_file_menu()
        file_menu.addSeparator()

        # File > Export...
        export_action = file_menu.addAction('E&xport...')
        export_action.setDisabled(True)  # Initially, all tracks are selected, so exporting is disabled.
        export_action.setShortcut(QKeySequence.StandardKey.Save)
        export_action.triggered.connect(self.tracks.export_selected_tracks)
        self._menu_actions[self.EXPORT_MENU_ACTION_KEY] = export_action

        # File > Close
        self.add_close_action_to_file_menu(file_menu)

        # Edit menu
        self._create_edit_menu()

        # Controls menu
        self._create_controls_menu()

        # Window menu
        self.create_window_menu()

        # Help menu
        self.create_help_menu()

    def _create_edit_menu(self) -> None:
        edit_menu = self.menu_bar.addMenu('&Edit')
        self.menu_bar.addAction(edit_menu.menuAction())

        for i, track in enumerate(self.tracks.tracks):
            # Edit > Mute/Unmute [Track]
            track_name = track.controls.name.text()
            mute_action = edit_menu.addAction(f'Mute {track_name}')
            mute_action.setShortcut(QKeySequence(f'Ctrl+M,Ctrl+{i+1}'))
            mute_action.triggered.connect(self.mute_menu_action(track))
            self._menu_actions[f'mute_{track_name}'] = mute_action

            # Edit > Solo/Unsolo [Track]
            solo_action = edit_menu.addAction(f'Solo {track_name}')
            solo_action.setShortcut(QKeySequence(f'Ctrl+L,Ctrl+{i+1}'))
            solo_action.triggered.connect(self.solo_menu_action(track))
            self._menu_actions[f'solo_{track_name}'] = solo_action

            if i < len(self.tracks.tracks) - 1:
                edit_menu.addSeparator()

    def _create_controls_menu(self) -> None:
        controls_menu = self.menu_bar.addMenu('&Controls')
        self.menu_bar.addAction(controls_menu.menuAction())

        # Controls > Play/Pause
        play_action = controls_menu.addAction('&Play')
        play_action.setShortcut(QKeySequence(Qt.Key.Key_Space))
        play_action.triggered.connect(self.play_menu_action)
        self._menu_actions[self.PLAY_MENU_ACTION_KEY] = play_action

        controls_menu.addSeparator()

        # Controls > Restart
        restart_action = controls_menu.addAction('&Restart')
        restart_action.setShortcut(QKeySequence('Ctrl+Left'))
        restart_action.triggered.connect(self.tracks.controls.restart_track)

        # Controls > Skip Back
        skip_back_action = controls_menu.addAction('Skip &Back')
        skip_back_action.setShortcut(QKeySequence(Qt.Key.Key_Left))
        skip_back_action.triggered.connect(self.tracks.controls.skip_back)

        # Controls > Skip Forward
        skip_forward_action = controls_menu.addAction('Skip &Forward')
        skip_forward_action.setShortcut(QKeySequence(Qt.Key.Key_Right))
        skip_forward_action.triggered.connect(self.tracks.controls.skip_forward)

        controls_menu.addSeparator()

        # Controls > Increase Volume
        increase_volume_action = controls_menu.addAction('&Increase Volume')
        increase_volume_action.setShortcut(QKeySequence('Ctrl+Up'))
        increase_volume_action.triggered.connect(self.tracks.controls.increase_volume)

        # Controls > Decrease Volume
        decrease_volume_action = controls_menu.addAction('&Decrease Volume')
        decrease_volume_action.setShortcut(QKeySequence('Ctrl+Down'))
        decrease_volume_action.triggered.connect(self.tracks.controls.decrease_volume)

    def output_file_path(self, input_file_paths: list[str]) -> str:
        name = '+'.join(os.path.splitext(os.path.basename(path))[0] for path in sorted(input_file_paths))
        _, extension = os.path.splitext(input_file_paths[0])
        return os.path.join(self.temp_dir_path, name + extension)
    
    def has_temp_mix_file(self, input_file_paths: list[str]) -> bool:
        if not input_file_paths:
            raise ValueError('No input files provided!')
        if len(input_file_paths) == 1 or (len(input_file_paths) == len(self.tracks.tracks) and self.source_file_path):
            return True
        return self.output_file_path(input_file_paths) in self._temp_file_paths

    # NOTE: This will create the temp mix file using ffmpeg if it does not exist already!
    def temp_file_path(self, input_file_paths: list[str]) -> str:
        if not input_file_paths:
            raise ValueError('No input files provided!')
        if len(input_file_paths) == 1:
            return input_file_paths[0]
        if len(input_file_paths) == len(self.tracks.tracks) and self.source_file_path:
            return self.source_file_path
        if (output_path := self.output_file_path(input_file_paths)) not in self._temp_file_paths:
            self.create_temp_mix_file(input_file_paths)
            self._temp_file_paths.add(output_path)
        return output_path

    def create_temp_mix_file(self, input_file_paths: list[str]) -> None:
        if not input_file_paths or len(input_file_paths) < 2:
            raise ValueError('Not enough input files provided!')
        output_path = self.output_file_path(input_file_paths)
        print(f'Creating temporary mix file "{output_path}"...')
        merge_audio_files(input_file_paths, output_path)

    def play_menu_action(self) -> None:
        self.tracks.controls.toggle_playback()
        self._menu_actions[self.PLAY_MENU_ACTION_KEY].setText('&Pause' if self.tracks.controls.playing else '&Play')

    def mute_menu_action(self, track: Track) -> Callable[[], None]:
        def handler():
            name = track.controls.name.text()
            track.controls.toggle_muted()
            self._menu_actions[f'mute_{name}'].setText(f'Unmute {name}' if track.muted else f'Mute {name}')
        return handler

    def solo_menu_action(self, track: Track) -> Callable[[], None]:
        def handler():
            prev_soloed_track = self.tracks.soloed_track
            name = track.controls.name.text()
            track.controls.toggle_soloed()
            self._menu_actions[f'solo_{name}'].setText(f'Unsolo {name}' if track.soloed else f'Solo {name}')
            self._menu_actions[f'mute_{name}'].setDisabled(track.soloed)
            if prev_soloed_track and prev_soloed_track != track:
                prev_name = prev_soloed_track.controls.name.text()
                self._menu_actions[f'solo_{prev_name}'].setText(f'Solo {prev_name}')
        return handler

    def enable_export_menu_action(self) -> None:
        self._menu_actions[self.EXPORT_MENU_ACTION_KEY].setDisabled(False)

    def disable_export_menu_action(self) -> None:
        self._menu_actions[self.EXPORT_MENU_ACTION_KEY].setDisabled(True)


class UnmixerUI:

    CHOOSE_TRACK_DIR_DIALOG_TITLE = 'Open Isolated Tracks Directory'

    MAX_RECENTLY_OPENED_SONGS = 10

    def __init__(self, song_path: Optional[str] = None, input_dir_path: Optional[str] = None) -> None:
        if song_path and input_dir_path:
            raise ValueError('Must provide song path or input directory path, but not both!')
        
        self.app = QApplication([])
        self.settings = QSettings(ORGANIZATION_NAME, APP_NAME)
        self.song_path = expand_path(song_path) if song_path else None
        self.input_dir_path = expand_path(input_dir_path) if input_dir_path else None
        self.output_dir_path = expand_path(self.setting(settings.importer.OUTPUT_DIR_PATH))
        self.output_format = self.setting(settings.importer.AUDIO_FORMAT)
        self.other_track_name = self.setting(settings.prefs.OTHER_TRACK_NAME)

        if self.should_disable_other_track_name_selection() and self.other_track_name != DEFAULT_OTHER_TRACK_NAME:
            self.update_setting(settings.prefs.OTHER_TRACK_NAME, DEFAULT_OTHER_TRACK_NAME)
            self.other_track_name = DEFAULT_OTHER_TRACK_NAME

        self.import_window = None
        self.preferences_window = None
        self.track_windows = {}
        self.active_window = None

        if self.input_dir_path:
            self.track_windows[self.input_dir_path] = UnmixerTrackExplorerWindow(self,
                                                                                 input_dir_path=self.input_dir_path)
            self.add_song_to_recently_opened()
        else:
            self.import_window = UnmixerImportWindow(self, source_file_path=self.song_path)

    def setting(self, key: str) -> Any:
        return self.settings.value(key, DEFAULT_SETTINGS[key])

    def update_setting(self, key: str, value: Any) -> None:
        self.settings.setValue(key, value)

    @staticmethod
    def open_project_readme() -> None:
        QDesktopServices.openUrl(QUrl(PROJECT_README_URL))

    def choose_track_directory(self) -> None:
        working_dir = expand_path(self.settings.value(
            settings.open.TRACK_DIR_PATH,
            self.setting(settings.importer.OUTPUT_DIR_PATH)
        ))
        input_path = QFileDialog.getExistingDirectory(self.active_window, self.CHOOSE_TRACK_DIR_DIALOG_TITLE,
                                                      working_dir, QFileDialog.Option.ShowDirsOnly)
        if not input_path:
            return

        self.update_setting(settings.open.TRACK_DIR_PATH, os.path.dirname(input_path))
        self.input_dir_path = input_path
        self.show_track_explorer_window(input_path)

    def run(self) -> None:
        active_window = None
        if self.track_windows:
            for window in self.track_windows.values():
                active_window = window
                window.show()
        if self.import_window:
            active_window = self.import_window
            self.import_window.show()
        self.active_window = active_window
        self.app.aboutToQuit.connect(self.stop_import_process)
        self.app.exec()

    def should_disable_other_track_name_selection(self) -> bool:
        return self.setting(settings.prefs.PRETRAINED_MODEL) == models.HTDEMUCS_6S
        
    def stop_import_process(self) -> None:
        if self.import_window and ((import_process := self.import_window.importer.import_process) and import_process.is_alive()):
            print(f'Attempting to terminate import process (PID {import_process.pid})...')
            import_process.terminate()
            cleanup_intermediate_dir(self.output_dir_path, self.setting(settings.prefs.PRETRAINED_MODEL))

    def add_song_to_recently_opened(self, dir_path: Optional[str] = None, reset_menu_bars: bool = True) -> None:
        dir_path = expand_path(dir_path or self.input_dir_path)
        if not dir_path:
            return
        
        recently_opened = self.settings.value(settings.open.RECENTLY_OPENED, [], 'QStringList')
        if dir_path in recently_opened:
            recently_opened.remove(dir_path)
        recently_opened.insert(0, dir_path)
        if len(recently_opened) > self.MAX_RECENTLY_OPENED_SONGS:
            recently_opened = recently_opened[:self.MAX_RECENTLY_OPENED_SONGS]
        
        self.update_setting(settings.open.RECENTLY_OPENED, recently_opened)
        self.settings.sync()
        if reset_menu_bars:
            self.reset_all_menu_bars()

    def import_finished(self, source_file_path: Optional[str] = None, input_dir_path: Optional[str] = None) -> None:
        # The isolated tracks are in a subdirectory of self.output_dir_path named for the song title.
        if source_file_path:
            song_title = os.path.splitext(os.path.basename(source_file_path))[0]
            if self.setting(settings.importer.CREATE_MODEL_SUBDIR):
                model_name = self.setting(settings.prefs.PRETRAINED_MODEL)
                dir_path = str(os.path.join(self.output_dir_path, model_name, song_title))
            else:
                dir_path = str(os.path.join(self.output_dir_path, song_title))
        else:
            dir_path = input_dir_path or self.input_dir_path
        dir_path = expand_path(dir_path)

        if self.setting(settings.prefs.SHOW_TRACK_EXPLORER_WHEN_IMPORT_FINISHED):
            self.show_track_explorer_window(dir_path, source_file_path=source_file_path)
        elif self.import_window and self.import_window.importer.import_process is None:
            self.import_window.reset()

        QMessageBox.information(self.active_window, SUCCESS_MESSAGE_TITLE,
                                f'Successfully wrote isolated tracks to {dir_path}.')

    def show_track_explorer_window(self, input_dir_path: str, source_file_path: Optional[str] = None) -> None:
        dir_path = expand_path(input_dir_path)
        if dir_path in self.track_windows:
            self.track_windows[dir_path].bring_to_front()
            return

        if not os.path.exists(dir_path):
            QMessageBox.warning(self.active_window, ERROR_MESSAGE_TITLE, f'Could not locate directory {dir_path}!')
            return

        window = UnmixerTrackExplorerWindow(self, input_dir_path=dir_path, source_file_path=source_file_path)
        window.show()
        self.track_windows[dir_path] = window
        self.add_song_to_recently_opened(dir_path, reset_menu_bars=False)

        if self.import_window:
            if self.import_window.importer.import_process is None:
                self.import_window.reset()
            else:
                self.import_window.reset_menu_bar()

        if self.preferences_window:
            self.preferences_window.reset_menu_bar()

        for window in self.track_windows.values():
            window.reset_menu_bar()

    def bring_import_window_to_front(self) -> None:
        if not self.import_window:
            self.import_window = UnmixerImportWindow(self)
            self.import_window.show()
        self.import_window.bring_to_front()

    def bring_preferences_window_to_front(self) -> None:
        if not self.preferences_window:
            self.preferences_window = UnmixerPreferencesWindow(self)
            self.preferences_window.show()
        self.preferences_window.bring_to_front()

    def reset_all_menu_bars(self) -> None:
        if self.import_window:
            self.import_window.reset_menu_bar()

        if self.preferences_window:
            self.preferences_window.reset_menu_bar()

        for window in self.track_windows.values():
            window.reset_menu_bar()



from typing import Optional
import os.path

from PyQt6.QtCore import Qt
from PyQt6.QtGui import QColor
from PyQt6.QtWidgets import QHBoxLayout, QLabel, QPushButton, QVBoxLayout, QWidget
from soundfile import SoundFile

from unmixer.ui.constants import FONT_WEIGHT_BOLD
from unmixer.ui.waveform import Waveform


class TrackControls(QWidget):
    
    DEFAULT_TRACK_NAME_FONT_SIZE = 15
    TRACK_NAME_FONT_SIZE_REDUCTION_AMOUNT = 3
    TRACK_NAME_REDUCTION_THRESHOLD = 10
    
    MIN_BUTTON_WIDTH = 85
    
    PLUS_SEPARATOR = ' + '
    
    MUTE_BUTTON_TEXT = '⬛️ Mute'
    UNMUTE_BUTTON_TEXT = '🟥 Unmute'
    
    SOLO_BUTTON_TEXT = '⬛️ Solo'
    UNSOLO_BUTTON_TEXT = '🟨 Unsolo'
    
    def __init__(self, sound_file: SoundFile, name: Optional[str] = None) -> None:
        super().__init__()
        tooltip = None
        if name is None:
            name, tooltip = self.format_track_name(sound_file.name)
        
        self.name = QLabel(name)
        font = self.name.font()
        font_size = self.DEFAULT_TRACK_NAME_FONT_SIZE
        if len(name) > self.TRACK_NAME_REDUCTION_THRESHOLD:
            font_size -= self.TRACK_NAME_FONT_SIZE_REDUCTION_AMOUNT * (name.count(self.PLUS_SEPARATOR) or 1)
        font.setPixelSize(font_size)
        font.setWeight(FONT_WEIGHT_BOLD)
        self.name.setFont(font)
        self.name.setAlignment(Qt.AlignmentFlag.AlignCenter)
        if tooltip:
            self.setStatusTip(tooltip)
            self.setToolTip(tooltip)
        
        self.muted = False
        self.mute_button = QPushButton(self.MUTE_BUTTON_TEXT)
        self.mute_button.setStyleSheet('text-align: left')
        self.mute_button.setMinimumWidth(self.MIN_BUTTON_WIDTH)
        self.mute_button.clicked.connect(self.toggle_muted)
        
        self.soloed = False
        self.solo_button = QPushButton(self.SOLO_BUTTON_TEXT)
        self.solo_button.setStyleSheet('text-align: left')
        self.solo_button.setMinimumWidth(self.MIN_BUTTON_WIDTH)
        self.solo_button.clicked.connect(self.toggle_soloed)
        
        layout = QVBoxLayout()
        layout.addWidget(self.name)
        layout.addWidget(self.mute_button)
        layout.addWidget(self.solo_button)
        self.setLayout(layout)
        self.show()
    
    @classmethod
    def format_track_name(cls, file_path: str) -> tuple[str, Optional[str]]:
        name, _ = os.path.splitext(os.path.basename(file_path))
        is_long_name = name.count('+') > 1
        words = []
        full_words = [] if is_long_name else None
        for word in name.split('+'):
            full_word = ' '.join(subword.capitalize() for subword in word.split('_'))
            if is_long_name:
                full_words.append(full_word)
            words.append(word[0].capitalize() if is_long_name else full_word)
        name = cls.PLUS_SEPARATOR.join(words)
        tooltip = cls.PLUS_SEPARATOR.join(full_words) if is_long_name else None
        return name, tooltip
    
    @property
    def track(self) -> 'Track':
        return self.parent()
    
    @property
    def soloed_track(self) -> Optional['Track']:
        return self.track.parent().soloed_track
    
    def toggle_muted_state(self) -> None:
        self.muted = not self.muted
        self.mute_button.setText(self.UNMUTE_BUTTON_TEXT if self.muted else self.MUTE_BUTTON_TEXT)
        self.update()
    
    def toggle_muted(self) -> None:
        if self.soloed:
            return
        self.toggle_muted_state()
        self.track.waveform.disabled = self.muted or (self.soloed_track is not None and self.soloed_track != self.track)
        if selected_count := len(self.track.parent().selected_tracks):
            self.track.parent().controls.enable_play_button()
            if selected_count == 1 or selected_count == len(self.track.parent().tracks):
                self.track.parent().controls.disable_export_button()
            else:
                self.track.parent().controls.enable_export_button()
        else:
            self.track.parent().controls.disable_export_button()
            self.track.parent().controls.disable_play_button()
            if self.track.parent().controls.playing:
                # Auto-pause if all tracks are muted.
                self.track.parent().pause_playback()
        self.track.parent().update_media_player_for_current_selected_tracks()
        self.track.parent().update()
    
    def toggle_soloed_state(self) -> None:
        self.soloed = not self.soloed
        self.solo_button.setText(self.UNSOLO_BUTTON_TEXT if self.soloed else self.SOLO_BUTTON_TEXT)
        if self.soloed and self.muted:
            self.toggle_muted_state()
        self.mute_button.setDisabled(self.soloed)
        self.update()
    
    def toggle_soloed(self) -> None:
        self.toggle_soloed_state()
        self.parent().parent().soloed_track = self.track if self.soloed else None
        if self.soloed:
            self.track.parent().controls.enable_play_button()
            self.track.parent().controls.disable_export_button()
        elif 1 < len(self.track.parent().selected_tracks) < len(self.track.parent().tracks):
            self.track.parent().controls.enable_export_button()
        self.track.parent().update_media_player_for_current_selected_tracks()
        self.track.parent().update()


class Track(QWidget):
    
    def __init__(self, sound_file: SoundFile, background_colors: Optional[tuple[QColor, QColor]] = None,
                 name: Optional[str] = None) -> None:
        super().__init__()
        self.file_path = sound_file.name
        self.controls = TrackControls(sound_file, name=name)
        self.waveform = Waveform(sound_file, background_colors)
        self.waveform.setStatusTip(self.controls.statusTip())
        self.waveform.setToolTip(self.controls.toolTip())
        
        layout = QHBoxLayout()
        layout.addWidget(self.controls, stretch=1)
        layout.addWidget(self.waveform, stretch=9)
        self.setLayout(layout)
        self.show()
    
    @property
    def muted(self) -> bool:
        return self.controls.muted
    
    @property
    def soloed(self) -> bool:
        return self.controls.soloed
    
    def enable_waveform(self) -> None:
        self.waveform.enable()
        
    def disable_waveform(self) -> None:
        self.waveform.disable()



from typing import Optional
import os.path
import shlex
import shutil

import demucs.separate
import ffmpeg

from unmixer.constants import (
    ALLOWED_CLIP_MODES,
    ALLOWED_WAV_BIT_DEPTHS,
    DEFAULT_CREATE_MODEL_SUBDIR,
    DEFAULT_ISOLATED_TRACK_FORMAT,
    DEFAULT_MP3_BITRATE_KBPS,
    DEFAULT_MP3_PRESET,
    DEFAULT_PRETRAINED_MODEL,
    DEFAULT_SPLIT_INPUT_INTO_SEGMENTS,
    DEFAULT_WAV_DEPTH,
    FLAC_FORMAT,
    ISOLATED_TRACK_FORMATS,
    MP3_FORMAT,
    MP3_PRESET_MAX,
    MP3_PRESET_MIN,
    OTHER_TRACK_NAME,
    WAV_FORMAT,
)
from unmixer.util import cleanup_intermediate_dir, expand_path, has_gpu_acceleration


MIX_FILTER_NAME = 'amix'


# Example: ffmpeg -i bass.wav -i drums.wav -i guitar.wav -filter_complex amix=inputs=3 -ac 2 instrumental.wav
def merge_audio_files(input_file_paths: list[str], output_file_path: str) -> None:
    if not input_file_paths or len(input_file_paths) < 2:
        raise ValueError('Not enough input files provided!')
    streams = [ffmpeg.input(expand_path(file_path)) for file_path in input_file_paths]
    ffmpeg.filter(streams, MIX_FILTER_NAME, inputs=len(streams), normalize=0). \
        output(expand_path(output_file_path), ac=2). \
        run()


# Reference: https://github.com/facebookresearch/demucs#calling-from-another-python-program
def create_isolated_tracks_from_audio_file(
        input_file_path: str, output_dir_path: str, *, remove_model_dir: Optional[bool] = None,
        other_track_name: Optional[str] = None, output_format: Optional[str] = None,
        model_name: Optional[str] = None, clip_mode: Optional[str] = None,
        disable_gpu_acceleration: Optional[bool] = None, cpu_parallelism: Optional[int] = None,
        mp3_preset: Optional[int] = None, mp3_bitrate_kbps: Optional[int] = None,
        wav_bit_depth: Optional[str] = None, split_into_segments: Optional[bool] = None,
        segment_length_seconds: Optional[int] = None, segment_overlap_percent: Optional[int] = None,
        random_shift_count: Optional[int] = None) -> None:
    input_file_path = expand_path(input_file_path)
    output_dir_path = expand_path(output_dir_path)

    if not output_format:
        output_format = DEFAULT_ISOLATED_TRACK_FORMAT
    if output_format not in ISOLATED_TRACK_FORMATS:
        raise ValueError(f'Unsupported output format "{output_format}"!')
    if output_format != MP3_FORMAT and (mp3_preset or mp3_bitrate_kbps):
        raise ValueError('MP3 preset and bitrate may only be provided when output format is MP3!')
    if output_format != WAV_FORMAT and wav_bit_depth:
        raise ValueError('WAV bit depth may only be provided when output format is WAV!')

    if clip_mode and clip_mode not in ALLOWED_CLIP_MODES:
        raise ValueError(f'Unsupported clip mode "{clip_mode}"!')

    if cpu_parallelism is not None:
        if has_gpu_acceleration() and not disable_gpu_acceleration:
            raise ValueError('CPU parallelism may only be provided when GPU acceleration is disabled!')
        if cpu_parallelism <= 0:
            raise ValueError('CPU parallelism must be a positive integer!')

    if mp3_preset and (mp3_preset < MP3_PRESET_MIN or mp3_preset > MP3_PRESET_MAX):
        raise ValueError(f'MP3 preset must be between {MP3_PRESET_MIN} and {MP3_PRESET_MAX}!')

    if mp3_bitrate_kbps and mp3_bitrate_kbps <= 0:
        raise ValueError('MP3 bitrate must be a positive integer!')

    if wav_bit_depth and wav_bit_depth not in ALLOWED_WAV_BIT_DEPTHS:
        raise ValueError(f'Unsupported WAV bit depth "{wav_bit_depth}"!')

    if split_into_segments is None:
        split_into_segments = DEFAULT_SPLIT_INPUT_INTO_SEGMENTS
    if split_into_segments:
        if segment_length_seconds and segment_length_seconds <= 0:
            raise ValueError('Segment length must be a positive integer!')
        if segment_overlap_percent and segment_overlap_percent < 0:
            raise ValueError('Segment overlap must be a nonnegative integer!')
        if random_shift_count and random_shift_count < 0:
            raise ValueError('Random shift count must be a nonnegative integer!')
    else:
        if segment_length_seconds:
            raise ValueError('Segment length may only be provided when input is split into segments!')
        if segment_overlap_percent:
            raise ValueError('Segment overlap may only be provided when input is split into segments!')
        if random_shift_count:
            raise ValueError('Random shift count may only be provided when input is split into segments!')

    print(f'Attempting to create isolated tracks from "{input_file_path}"...')
    
    if not os.path.exists(output_dir_path):
        os.makedirs(output_dir_path)
    
    demucs_args = [input_file_path, '--out', output_dir_path]
    if model_name:
        demucs_args.extend(['--name', model_name])
    if clip_mode:
        demucs_args.extend(['--clip-mode', clip_mode])

    if output_format == FLAC_FORMAT:
        demucs_args.append('--flac')
    elif output_format == MP3_FORMAT:
        mp3_preset = mp3_preset or DEFAULT_MP3_PRESET
        mp3_bitrate_kbps = mp3_bitrate_kbps or DEFAULT_MP3_BITRATE_KBPS
        demucs_args.extend(['--mp3', '--mp3-preset', str(mp3_preset), '--mp3-bitrate', str(mp3_bitrate_kbps)])
    elif wav_bit_depth and wav_bit_depth != DEFAULT_WAV_DEPTH:
        demucs_args.append(f'--{wav_bit_depth}')

    if split_into_segments:
        if segment_length_seconds:
            demucs_args.extend(['--segment', str(segment_length_seconds)])
        if segment_overlap_percent:
            demucs_args.extend(['--overlap', f'{(segment_overlap_percent / 100):0.2f}'])
        if random_shift_count:
            demucs_args.extend(['--shifts', str(random_shift_count)])
    else:
        demucs_args.append('--no-split')

    if disable_gpu_acceleration:
        demucs_args.extend(['--device', 'cpu'])
    if cpu_parallelism and (not has_gpu_acceleration() or disable_gpu_acceleration):
        demucs_args.extend(['--jobs', str(cpu_parallelism)])

    print(f'Invoking demucs with args: {shlex.join(demucs_args)}')
    demucs.separate.main(demucs_args)
    
    # demucs creates an intermediate directory for the model name.
    # Move the isolated tracks out of this directory if desired.
    if remove_model_dir is None:
        remove_model_dir = not DEFAULT_CREATE_MODEL_SUBDIR
    file_name = os.path.splitext(os.path.basename(input_file_path))[0]
    model_name = model_name or DEFAULT_PRETRAINED_MODEL

    if remove_model_dir:
        output_dir = os.path.join(output_dir_path, file_name)
        shutil.move(os.path.join(output_dir_path, model_name, file_name), output_dir)
        cleanup_intermediate_dir(output_dir_path, model_name)
    else:
        output_dir = os.path.join(output_dir_path, model_name, file_name)

    print(f'Wrote isolated tracks to "{output_dir}".')
    
    if other_track_name and (other_track := next((n for n in os.listdir(output_dir) if n.startswith(OTHER_TRACK_NAME)), None)):
        _, extension = os.path.splitext(other_track)
        shutil.move(os.path.join(output_dir, other_track), os.path.join(output_dir, other_track_name + extension))



from multiprocessing import Process
from typing import Optional
import os

from PyQt6.QtCore import Qt, QTimer
from PyQt6.QtWidgets import (
    QButtonGroup,
    QFileDialog,
    QFormLayout,
    QHBoxLayout,
    QLabel,
    QMessageBox,
    QPushButton,
    QRadioButton,
    QVBoxLayout,
    QWidget,
)

from unmixer.constants import (
    ALLOWED_OTHER_TRACK_NAMES,
    ISOLATED_TRACK_FORMATS,
    MP3_FORMAT,
    WAV_FORMAT,
    settings,
)
from unmixer.remix import create_isolated_tracks_from_audio_file
from unmixer.ui.constants import (
    ERROR_MESSAGE_TITLE,
    FONT_WEIGHT_BOLD,
    OTHER_TRACK_NAME_DISABLED_TOOLTIP,
)
from unmixer.util import expand_path, has_gpu_acceleration


class SongImportSettings(QWidget):

    AUDIO_FORMAT_LABEL = 'Isolated track output format'
    OTHER_TRACK_LABEL = '"Other" track name'

    OTHER_SETTINGS_BUTTON_TEXT = '⚙️ More settings...'
    OTHER_SETTINGS_BUTTON_WIDTH = 200

    LABEL_SPACING = 20

    def __init__(self, app: 'UnmixerUI') -> None:
        super().__init__()
        self.app = app

        form_layout = QFormLayout()
        form_layout.setHorizontalSpacing(self.LABEL_SPACING)

        self.audio_format_label = QLabel(self.AUDIO_FORMAT_LABEL)
        font = self.audio_format_label.font()
        font.setWeight(FONT_WEIGHT_BOLD)
        self.audio_format_label.setFont(font)

        self.audio_format_button_group = QButtonGroup()
        self.audio_format_button_group.setExclusive(True)
        self.audio_format_button_group.buttonToggled.connect(self.update_audio_format)

        audio_format_button_layout = QHBoxLayout()
        for audio_format in sorted(list(ISOLATED_TRACK_FORMATS)):
            format_button = QRadioButton(audio_format.upper())
            if audio_format == self.app.output_format:
                format_button.setChecked(True)
            audio_format_button_layout.addWidget(format_button)
            self.audio_format_button_group.addButton(format_button)

        self.other_track_label = QLabel(self.OTHER_TRACK_LABEL)
        font = self.other_track_label.font()
        font.setWeight(FONT_WEIGHT_BOLD)
        self.other_track_label.setFont(font)

        self.other_track_button_group = QButtonGroup()
        self.other_track_button_group.setExclusive(True)
        self.other_track_button_group.buttonToggled.connect(self.update_other_track_name)

        other_track_button_layout = QHBoxLayout()
        disable_other_track_buttons = self.app.should_disable_other_track_name_selection()
        for other_name in ALLOWED_OTHER_TRACK_NAMES:
            name_button = QRadioButton(other_name.capitalize())
            if other_name == self.app.other_track_name:
                name_button.setChecked(True)
            if disable_other_track_buttons:
                name_button.setDisabled(True)
                name_button.setToolTip(OTHER_TRACK_NAME_DISABLED_TOOLTIP)
            other_track_button_layout.addWidget(name_button)
            self.other_track_button_group.addButton(name_button)

        form_layout.addRow(self.audio_format_label, audio_format_button_layout)
        form_layout.addRow(self.other_track_label, other_track_button_layout)

        self.other_settings_button = QPushButton(self.OTHER_SETTINGS_BUTTON_TEXT)
        self.other_settings_button.setFixedWidth(self.OTHER_SETTINGS_BUTTON_WIDTH)
        self.other_settings_button.clicked.connect(self.app.bring_preferences_window_to_front)

        layout = QVBoxLayout()
        layout.addLayout(form_layout)
        sublayout = QHBoxLayout()
        sublayout.addWidget(self.other_settings_button)
        layout.addLayout(sublayout)
        self.setLayout(layout)
        self.show()

    def set_audio_format(self, audio_format: str) -> None:
        audio_format = audio_format.lower()
        if audio_format not in ISOLATED_TRACK_FORMATS:
            raise ValueError(f'Unsupported audio format "{audio_format}"!')

        for button in self.audio_format_button_group.buttons():
            button.setChecked(button.text().lower() == audio_format)

    def update_audio_format(self, button: QRadioButton, checked: bool) -> None:
        if checked and (audio_format := button.text().lower()) != self.app.output_format:
            self.app.output_format = audio_format
            self.app.update_setting(settings.importer.AUDIO_FORMAT, self.app.output_format)
            if self.app.preferences_window:
                self.app.preferences_window.set_audio_format(self.app.output_format)

    def set_other_track_name(self, other_track_name: str) -> None:
        other_track_name = other_track_name.lower()
        if other_track_name not in ALLOWED_OTHER_TRACK_NAMES:
            raise ValueError(f'Unsupported track name "{other_track_name}"!')

        self.app.other_track_name = other_track_name
        for button in self.other_track_button_group.buttons():
            button.setChecked(button.text().lower() == self.app.other_track_name)

    def update_other_track_name(self, button: QRadioButton, checked: bool) -> None:
        if checked and (track_name := button.text().lower()) != self.app.other_track_name:
            self.app.other_track_name = track_name
            self.app.update_setting(settings.prefs.OTHER_TRACK_NAME, self.app.other_track_name)
            if self.app.preferences_window:
                self.app.preferences_window.set_other_track_name(self.app.other_track_name)

    def disable_other_track_name_buttons(self) -> None:
        for button in self.other_track_button_group.buttons():
            button.setDisabled(True)
            button.setToolTip(OTHER_TRACK_NAME_DISABLED_TOOLTIP)

    def enable_other_track_name_buttons(self) -> None:
        for button in self.other_track_button_group.buttons():
            button.setDisabled(False)
            button.setToolTip('')

    def disable_form(self) -> None:
        for button in self.audio_format_button_group.buttons():
            button.setDisabled(True)
        self.disable_other_track_name_buttons()

    def enable_form(self) -> None:
        for button in self.audio_format_button_group.buttons():
            button.setDisabled(False)
        if not self.app.should_disable_other_track_name_selection():
            self.enable_other_track_name_buttons()


class SongImporter(QWidget):
    
    EXPLORE_BUTTON_TEXT = '🎛️ Explore isolated tracks...'
    EXPLORE_BUTTON_TOOLTIP = 'Select a directory containing previously isolated tracks to explore.'

    CHOOSE_BUTTON_TEXT = '📂 Choose a song...'
    CHOOSE_BUTON_TOOLTIP = 'Select a new song to create isolated tracks from.'
    CHOOSE_DIFFERENT_SONG_BUTTON_TEXT = '📂 Choose a different song...'
    CHOOSE_DIALOG_TITLE = 'Select a Song'

    START_BUTTON_TEXT = '✅ Start'
    CANCEL_BUTTON_TEXT = '❌ Cancel'
    
    DEFAULT_TITLE_TEXT = 'Select a song to unmix'
    IMPORTING_TITLE_TEXT_TEMPLATE = 'Now creating isolated tracks from:\n{filename}'
    SONG_SELECTED_TITLE_TEXT_TEMPLATE = 'Selected song:\n{filename}'

    DEFAULT_SUBTITLE_TEXT = 'You can also drop a song here from your computer.'
    IMPORTING_SUBTITLE_TEXT = (
        'Please wait; this may take some time.\n'
        'The time required depends on the length of the song and the current prediction model.'
    )
    SONG_SELECTED_SUBTITLE_TEXT = f'Press [{START_BUTTON_TEXT}] to begin unmixing the song into isolated tracks.'
    
    IMPORT_STATUS_TIMER_INTERVAL_MILLIS = 1000
    
    MIN_BUTTON_HEIGHT = 50
    
    TITLE_FONT_SIZE = 20
    
    def __init__(self, app: 'UnmixerUI', input_file_path: Optional[str] = None) -> None:
        super().__init__()
        self.app = app
        self.input_file_path = expand_path(input_file_path) if input_file_path else None
        
        self._import_process = None
        self.check_import_status_timer = QTimer(self)
        self.check_import_status_timer.setInterval(self.IMPORT_STATUS_TIMER_INTERVAL_MILLIS)
        self.check_import_status_timer.timeout.connect(self.check_import_status)
        
        self.title = QLabel(self.DEFAULT_TITLE_TEXT)
        font = self.title.font()
        font.setPixelSize(self.TITLE_FONT_SIZE)
        font.setWeight(FONT_WEIGHT_BOLD)
        self.title.setFont(font)
        self.title.setAlignment(Qt.AlignmentFlag.AlignCenter)
        
        self.subtitle = QLabel(self.DEFAULT_SUBTITLE_TEXT)
        self.subtitle.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.subtitle.setMinimumHeight(self.MIN_BUTTON_HEIGHT * 2)
        
        self.import_settings = SongImportSettings(app)

        self.explore_tracks_button = QPushButton(self.EXPLORE_BUTTON_TEXT)
        self.explore_tracks_button.setMinimumHeight(self.MIN_BUTTON_HEIGHT)
        self.explore_tracks_button.setToolTip(self.EXPLORE_BUTTON_TOOLTIP)
        self.explore_tracks_button.clicked.connect(self.app.choose_track_directory)

        self.choose_file_button = QPushButton(self.CHOOSE_BUTTON_TEXT)
        self.choose_file_button.setMinimumHeight(self.MIN_BUTTON_HEIGHT)
        self.choose_file_button.setToolTip(self.CHOOSE_BUTON_TOOLTIP)
        self.choose_file_button.clicked.connect(self.choose_file)
        
        self.start_import_button = QPushButton(self.START_BUTTON_TEXT)
        self.start_import_button.setAutoDefault(True)
        self.start_import_button.setDefault(True)
        self.start_import_button.setHidden(True)
        self.start_import_button.setMinimumHeight(self.MIN_BUTTON_HEIGHT)
        self.start_import_button.clicked.connect(self.start_import_button_pressed)

        layout = QVBoxLayout()
        layout.addWidget(self.title)
        layout.addWidget(self.subtitle)
        layout.addWidget(self.import_settings)
        layout.addWidget(self.explore_tracks_button)
        sublayout = QHBoxLayout()
        sublayout.addWidget(self.choose_file_button)
        sublayout.addWidget(self.start_import_button)
        layout.addLayout(sublayout)
        self.setLayout(layout)

        self.setFocusProxy(self.choose_file_button)
        self.show()
        
        if self.input_file_path:
            self.select_song()
        else:
            self.import_settings.disable_form()
            self.import_settings.setHidden(True)
    
    @property
    def import_process(self) -> Optional[Process]:
        return self._import_process
    
    def check_import_status(self) -> None:
        # Wait for import process to return/exit, then replace the import window with the track explorer.
        if self._import_process and not self._import_process.is_alive():
            self.check_import_status_timer.stop()
            return_code = self._import_process.exitcode
            self._import_process = None
            if return_code == 0:
                self.parent().app.import_finished(source_file_path=self.input_file_path,
                                                  input_dir_path=self.app.output_dir_path)
            else:
                self.select_song()
                QMessageBox.critical(self.parent(), ERROR_MESSAGE_TITLE,
                                     f'An error occurred while unmixing {os.path.basename(self.input_file_path)}.')
    
    def choose_file(self) -> None:
        working_dir = self.parent().app.setting(settings.importer.DIR_PATH)
        input_path, _ = QFileDialog.getOpenFileName(self, self.CHOOSE_DIALOG_TITLE, working_dir)
        if not input_path:
            return
        
        self.input_file_path = input_path
        self.parent().app.update_setting(settings.importer.DIR_PATH, os.path.dirname(input_path))
        self.select_song()

    def select_song(self) -> None:
        if not self.input_file_path:
            return

        if self.parent():
            self.parent().enable_import_menu_action()
            self.parent().setWindowTitle(self.parent().DEFAULT_WINDOW_TITLE)
            self.parent().update()

        filename = os.path.basename(self.input_file_path)
        self.title.setText(self.SONG_SELECTED_TITLE_TEXT_TEMPLATE.format(filename=filename))
        self.title.setToolTip(self.input_file_path)
        self.subtitle.setText(self.SONG_SELECTED_SUBTITLE_TEXT)
        self.import_settings.enable_form()
        self.import_settings.setHidden(False)
        self.choose_file_button.setDisabled(False)
        self.choose_file_button.setText(self.CHOOSE_DIFFERENT_SONG_BUTTON_TEXT)
        self.start_import_button.setHidden(False)
        self.start_import_button.setText(self.START_BUTTON_TEXT)
        self.start_import_button.setFocus()
        self.update()

    def start_import_button_pressed(self) -> None:
        if self._import_process is not None and self._import_process.is_alive():
            self.app.stop_import_process()
            self._import_process = None
            self.check_import_status_timer.stop()
            self.select_song()
        else:
            self.import_file()

    def import_file(self) -> None:
        if not self.input_file_path:
            return
        
        if self.parent():
            self.parent().disable_import_menu_action()
            self.parent().setWindowTitle(self.parent().IMPORTING_WINDOW_TITLE)
            self.parent().update()

        filename = os.path.basename(self.input_file_path)
        self.title.setText(self.IMPORTING_TITLE_TEXT_TEMPLATE.format(filename=filename))
        self.title.setToolTip(self.input_file_path)
        self.subtitle.setText(self.IMPORTING_SUBTITLE_TEXT)
        self.import_settings.disable_form()
        self.choose_file_button.setDisabled(True)
        self.start_import_button.setText(self.CANCEL_BUTTON_TEXT)
        self.update()

        import_kwargs = {
            'clip_mode': self.app.setting(settings.prefs.CLIP_MODE),
            'disable_gpu_acceleration': self.app.setting(settings.prefs.DISABLE_GPU_ACCELERATION),
            'model_name': self.app.setting(settings.prefs.PRETRAINED_MODEL),
            'other_track_name': self.app.other_track_name,
            'output_format': self.app.output_format,
            'remove_model_dir': not self.app.setting(settings.importer.CREATE_MODEL_SUBDIR),
            'split_into_segments': self.app.setting(settings.prefs.SPLIT_INTO_SEGMENTS),
        }
        if not has_gpu_acceleration() or import_kwargs['disable_gpu_acceleration']:
            import_kwargs['cpu_parallelism'] = self.app.setting(settings.prefs.CPU_PARALLELISM)
        if import_kwargs['split_into_segments']:
            import_kwargs['segment_length_seconds'] = self.app.setting(settings.prefs.SEGMENT_LENGTH)
            import_kwargs['segment_overlap_percent'] = self.app.setting(settings.prefs.SEGMENT_OVERLAP)
            import_kwargs['random_shift_count'] = self.app.setting(settings.prefs.SHIFT_COUNT)
        if self.app.output_format == MP3_FORMAT:
            import_kwargs['mp3_preset'] = self.app.setting(settings.prefs.MP3_PRESET)
            import_kwargs['mp3_bitrate_kbps'] = self.app.setting(settings.prefs.MP3_BITRATE)
        elif self.app.output_format == WAV_FORMAT:
            import_kwargs['wav_bit_depth'] = self.app.setting(settings.prefs.WAV_BIT_DEPTH)

        self._import_process = Process(target=create_isolated_tracks_from_audio_file,
                                       args=(self.input_file_path, self.app.output_dir_path), kwargs=import_kwargs)
        self._import_process.start()
        self.check_import_status_timer.start()

    def reset(self) -> None:
        if self._import_process is not None and self._import_process.is_alive():
            self.app.stop_import_process()
            self.check_import_status_timer.stop()

        self._import_process = None
        self.input_file_path = None
        self.title.setText(self.DEFAULT_TITLE_TEXT)
        self.title.setToolTip('')
        self.subtitle.setText(self.DEFAULT_SUBTITLE_TEXT)
        self.import_settings.disable_form()
        self.import_settings.setHidden(True)
        self.choose_file_button.setText(self.CHOOSE_BUTTON_TEXT)
        self.choose_file_button.setDisabled(False)
        self.start_import_button.setText(self.START_BUTTON_TEXT)
        self.start_import_button.setHidden(True)
        self.update()



import os.path
import shutil

from demucs.pretrained import _parse_remote_files, REMOTE_ROOT
from demucs.repo import BagOnlyRepo, RemoteRepo
import torch

from unmixer.constants import DEFAULT_ALLOWED_PRETRAINED_MODELS, ISOLATED_TRACK_FORMATS


def is_isolated_track(file_path: str) -> bool:
    return os.path.splitext(file_path)[1].lower().lstrip('.') in ISOLATED_TRACK_FORMATS


def cleanup_intermediate_dir(output_dir_path: str, model_name: str) -> None:
    intermediate_dir = os.path.join(output_dir_path, model_name)
    if os.path.exists(intermediate_dir) and not os.listdir(intermediate_dir):
        print(f'Removing intermediate directory {intermediate_dir}.')
        shutil.rmtree(intermediate_dir)


def expand_path(path: str) -> str:
    return os.path.abspath(os.path.expanduser(path))


# NOTE: This may break in a future version of demucs!
# Switch to using `demucs --list-models` as soon as it is available!
def get_available_pretrained_models() -> list[str]:
    try:
        models = _parse_remote_files(REMOTE_ROOT / 'files.txt')
        model_repo = RemoteRepo(models)
        bag_repo = BagOnlyRepo(REMOTE_ROOT, model_repo)
        model_names = bag_repo._bags.keys()
    except Exception as e:
        print(f'Failed to load latest pretrained models from demucs: {e}')
        model_names = DEFAULT_ALLOWED_PRETRAINED_MODELS
    return sorted(model_names)


def has_gpu_acceleration() -> bool:
    return torch.cuda.is_available()



from types import SimpleNamespace
import math
import os

from demucs.pretrained import DEFAULT_MODEL


DEFAULT_OUTPUT_DIR = '~/unmixer'

GUITAR_TRACK_NAME = 'guitar'
OTHER_TRACK_NAME = 'other'
PIANO_TRACK_NAME = 'piano'
DEFAULT_OTHER_TRACK_NAME = OTHER_TRACK_NAME
ALLOWED_OTHER_TRACK_NAMES = (GUITAR_TRACK_NAME, PIANO_TRACK_NAME, OTHER_TRACK_NAME)  # in display order

FLAC_FORMAT = 'flac'
MP3_FORMAT = 'mp3'
WAV_FORMAT = 'wav'
DEFAULT_ISOLATED_TRACK_FORMAT = WAV_FORMAT
ISOLATED_TRACK_FORMATS = {FLAC_FORMAT, MP3_FORMAT, WAV_FORMAT}

MIN_VOLUME = 0
MAX_VOLUME = 100
DEFAULT_VOLUME = 80

MP3_PRESET_MIN = 2  # Better quality
MP3_PRESET_MAX = 7  # Faster encoding
DEFAULT_MP3_PRESET = MP3_PRESET_MIN

# Reference: https://en.wikipedia.org/wiki/MP3#Bit_rate
ALLOWED_MP3_BITRATES_KBPS = (96, 112, 128, 160, 192, 224, 256, 320)
DEFAULT_MP3_BITRATE_KBPS = 320

WAV_DEPTH_INT16 = 'int16'
WAV_DEPTH_INT24 = 'int24'
WAV_DEPTH_FLOAT32 = 'float32'
DEFAULT_WAV_DEPTH = WAV_DEPTH_INT16
ALLOWED_WAV_BIT_DEPTHS = (WAV_DEPTH_INT16, WAV_DEPTH_INT24, WAV_DEPTH_FLOAT32)  # in display order

DEFAULT_CREATE_MODEL_SUBDIR = True
DEFAULT_DISABLE_GPU_ACCELERATION = False
DEFAULT_SHOW_TRACK_EXPLORER_WHEN_IMPORT_FINISHED = True
DEFAULT_SPLIT_INPUT_INTO_SEGMENTS = True

CLAMP_CLIP_MODE = 'clamp'
RESCALE_CLIP_MODE = 'rescale'
DEFAULT_CLIP_MODE = RESCALE_CLIP_MODE
ALLOWED_CLIP_MODES = (CLAMP_CLIP_MODE, DEFAULT_CLIP_MODE)

MIN_CPU_PARALLELISM = 1
DEFAULT_CPU_PARALLELISM = MIN_CPU_PARALLELISM

DEFAULT_SEGMENT_LENGTH_SECONDS = 10
MIN_SEGMENT_LENGTH_SECONDS = 1
DEFAULT_MAX_SEGMENT_LENGTH_SECONDS = 120
# Output from demucs with htdemucs-based models and segment lengths > 7:
# FATAL: Cannot use a Transformer model with a longer segment than it was trained for. Maximum segment is: 7.8
SHORT_MAX_SEGMENT_LENGTH_SECONDS = math.floor(7.8)


MIN_SEGMENT_OVERLAP_PERCENT = 0
MAX_SEGMENT_OVERLAP_PERCENT = 50
DEFAULT_SEGMENT_OVERLAP_PERCENT = 25

MIN_SHIFT_COUNT = 0
MAX_SHIFT_COUNT = 10
DEFAULT_SHIFT_COUNT = MIN_SHIFT_COUNT


######################
# Pre-trained Models #
######################

DEFAULT_PRETRAINED_MODEL = DEFAULT_MODEL

models = SimpleNamespace()
models.HDEMUCS_MMI = 'hdemucs_mmi'
models.HTDEMUCS = 'htdemucs'
models.HTDEMUCS_6S = 'htdemucs_6s'
models.HTDEMUCS_FT = 'htdemucs_ft'
models.MDX = 'mdx'
models.MDX_EXTRA = 'mdx_extra'
models.MDX_EXTRA_Q = 'mdx_extra_q'
models.MDX_Q = 'mdx_q'

DEFAULT_ALLOWED_PRETRAINED_MODELS = (
    models.HDEMUCS_MMI,
    models.HTDEMUCS,
    models.HTDEMUCS_6S,
    models.HTDEMUCS_FT,
    models.MDX,
    models.MDX_EXTRA,
    models.MDX_EXTRA_Q,
    models.MDX_Q,
)


#################
# Settings Keys #
#################

settings = SimpleNamespace()

# -- Import Settings -- #
settings.importer = SimpleNamespace()
# The audio format used when importing songs (str).
settings.importer.AUDIO_FORMAT = 'import/audioFormat'
# Whether or not to create a subdirectory of OUTPUT_DIR_PATH named for the prediction model (bool).
settings.importer.CREATE_MODEL_SUBDIR = 'import/createModelSubdir'
# The working directory for the file dialog presented when importing a song (str).
settings.importer.DIR_PATH = 'import/dirPath'
# The output directory for isolated tracks created when importing songs (str).
settings.importer.OUTPUT_DIR_PATH = 'import/outputDirPath'

# -- Open File/Directory Settings -- #
settings.open = SimpleNamespace()
# The most recently opened isolated track directories, used to populate the File > Recently Opened menu (list[str]).
settings.open.RECENTLY_OPENED = 'open/recentlyOpened'
# The working directory for the file dialog presented when opening existing isolated tracks (str).
settings.open.TRACK_DIR_PATH = 'open/trackDirPath'

# -- Playback Settings -- #
settings.playback = SimpleNamespace()
# The most recent playback volume (int).
settings.playback.VOLUME = 'playback/volume'

# -- User Preferences -- #
settings.prefs = SimpleNamespace()
# The strategy to use for avoiding clipping (str). Valid values are 'clamp' and 'rescale'.
settings.prefs.CLIP_MODE = 'prefs/clipMode'
# Number of parallel jobs to use when running on CPU (int).
settings.prefs.CPU_PARALLELISM = 'prefs/cpuParallelism'
# Whether or not to disable GPU acceleration (i.e., force running on CPU) for demucs (bool).
settings.prefs.DISABLE_GPU_ACCELERATION = 'prefs/disableGPUAcceleration'
# MP3 bitrate to use when MP3 format is selected (int). Value is in kbps.
settings.prefs.MP3_BITRATE = 'prefs/mp3BitRate'
# MP3 preset to use when MP3 format is selected (int). Range is 2 (better quality) to 7 (faster encoding).
settings.prefs.MP3_PRESET = 'prefs/mp3Preset'
# The name of the "other" track, used for importing songs and exploring isolated tracks (str).
settings.prefs.OTHER_TRACK_NAME = 'prefs/otherTrackName'
# The name of the pre-trained model to use for source separation (str).
settings.prefs.PRETRAINED_MODEL = 'prefs/pretrainedModel'
# Segment length when song is split into segments for processing (int). Value is in seconds.
settings.prefs.SEGMENT_LENGTH = 'prefs/segmentLength'
# Segment overlap when song is split into segments for processing (int). Value is a percentage (0-100).
settings.prefs.SEGMENT_OVERLAP = 'prefs/segmentOverlap'
# Number of random shifts to perform during prediction (int).
settings.prefs.SHIFT_COUNT = 'prefs/shiftCount'
# Whether or not to show the Track Explorer window after importing a song (bool).
settings.prefs.SHOW_TRACK_EXPLORER_WHEN_IMPORT_FINISHED = 'prefs/showTrackExplorer'
# Whether or not to split input songs into segments for processing (bool).
settings.prefs.SPLIT_INTO_SEGMENTS = 'prefs/splitIntoSegments'
# Bit depth to use when WAV format is selected (str). Valid values are 'int16', 'int24', and 'float32'.
settings.prefs.WAV_BIT_DEPTH = 'prefs/wavBitDepth'


####################
# Default Settings #
####################

DEFAULT_SETTINGS = {
    settings.importer.AUDIO_FORMAT: DEFAULT_ISOLATED_TRACK_FORMAT,
    settings.importer.CREATE_MODEL_SUBDIR: DEFAULT_CREATE_MODEL_SUBDIR,
    settings.importer.DIR_PATH: os.getcwd(),
    settings.importer.OUTPUT_DIR_PATH: DEFAULT_OUTPUT_DIR,
    settings.open.RECENTLY_OPENED: [],
    settings.open.TRACK_DIR_PATH: None,  # Default is the current output dir path setting.
    settings.playback.VOLUME: DEFAULT_VOLUME,
    settings.prefs.CLIP_MODE: DEFAULT_CLIP_MODE,
    settings.prefs.CPU_PARALLELISM: DEFAULT_CPU_PARALLELISM,
    settings.prefs.DISABLE_GPU_ACCELERATION: DEFAULT_DISABLE_GPU_ACCELERATION,
    settings.prefs.MP3_BITRATE: DEFAULT_MP3_BITRATE_KBPS,
    settings.prefs.MP3_PRESET: DEFAULT_MP3_PRESET,
    settings.prefs.OTHER_TRACK_NAME: DEFAULT_OTHER_TRACK_NAME,
    settings.prefs.PRETRAINED_MODEL: DEFAULT_PRETRAINED_MODEL,
    settings.prefs.SEGMENT_LENGTH: DEFAULT_SEGMENT_LENGTH_SECONDS,
    settings.prefs.SEGMENT_OVERLAP: DEFAULT_SEGMENT_OVERLAP_PERCENT,
    settings.prefs.SHIFT_COUNT: DEFAULT_SHIFT_COUNT,
    settings.prefs.SHOW_TRACK_EXPLORER_WHEN_IMPORT_FINISHED: DEFAULT_SHOW_TRACK_EXPLORER_WHEN_IMPORT_FINISHED,
    settings.prefs.SPLIT_INTO_SEGMENTS: DEFAULT_SPLIT_INPUT_INTO_SEGMENTS,
    settings.prefs.WAV_BIT_DEPTH: DEFAULT_WAV_DEPTH,
}



from typing import Optional
import random

from PyQt6.QtCore import QLineF, QRectF, Qt
from PyQt6.QtGui import QColor, QGradient, QLinearGradient, QPainter, QPainterPath
from PyQt6.QtWidgets import QWidget
from soundfile import SoundFile
import numpy as np


WAVEFORM_BACKGROUND_COLORS = [
    (QColor(20, 80, 200), QColor(90, 180, 240)),    # blue   (bass)
    (QColor(5, 175, 30), QColor(100, 240, 120)),    # green  (drums)
    (QColor(230, 30, 15), QColor(240, 110, 100)),   # red    (other/guitar/piano)
    (QColor(245, 225, 10), QColor(235, 240, 150)),  # yellow (vocals)
    (QColor(245, 90, 10), QColor(250, 150, 100)),   # orange
    (QColor(100, 5, 175), QColor(155, 55, 245)),    # purple
]


# Reference: https://gist.github.com/SpotlightKid/33ed933944beed851697039142613d98
class Waveform(QWidget):
    
    FOREGROUND_COLOR = QColor('white')
    WAVEFORM_COLOR = QColor(40, 40, 40)  # dark gray
    DISABLED_WAVEFORM_OVERLAY_COLOR = QColor(200, 200, 200, 200)  # transparent gray

    def __init__(self, sound_file: SoundFile, background_colors: Optional[tuple[QColor, QColor]] = None) -> None:
        super().__init__()
        if background_colors is None:
            background_colors = random.choice(WAVEFORM_BACKGROUND_COLORS)
        self.disabled = False
        self.frames = sound_file.read()
        self.channels = sound_file.channels
        self.sample_rate = sound_file.samplerate

        self.background_gradient = QLinearGradient()
        self.background_gradient.setColorAt(0, background_colors[0])
        self.background_gradient.setColorAt(1, background_colors[1])
        self.background_gradient.setSpread(QGradient.Spread.ReflectSpread)

    def paintEvent(self, event) -> None:
        painter = QPainter()
        painter.begin(self)
        self.draw_waveform(painter)
        painter.end()

    def draw_waveform(self, painter: QPainter) -> None:
        painter.setRenderHint(QPainter.RenderHint.Antialiasing)
        height = painter.device().height()
        width = painter.device().width()
        x_axis = height / 2
        samples_per_pixel = len(self.frames) / width

        # Draw background.
        self.background_gradient.setStart(0, x_axis)
        self.background_gradient.setFinalStop(0, 0)
        path = QPainterPath()
        path.addRoundedRect(QRectF(0, 0, width, height), 10, 10)
        pen = painter.pen()
        pen.setColor(self.FOREGROUND_COLOR)
        pen.setWidth(2)
        painter.setPen(pen)
        painter.fillPath(path, self.background_gradient)
        painter.drawPath(path)

        # Draw waveform.
        pen.setColor(self.WAVEFORM_COLOR)
        pen.setWidth(1)
        painter.setPen(pen)
        for x in range(width):
            if 0 <= (start := round(x * samples_per_pixel)) < len(self.frames):
                end = max(start + 1, start + int(samples_per_pixel))
                values = self.frames[start:end]
                if (max_value := np.max(values)) > 0:
                    y = x_axis - (x_axis * max_value)
                    painter.drawLine(QLineF(x, x_axis, x, y))
                if (min_value := np.min(values)) < 0:
                    y = x_axis - (x_axis * min_value)
                    painter.drawLine(QLineF(x, x_axis, x, y))

        # Draw center line (x-axis).
        pen.setColor(self.FOREGROUND_COLOR)
        pen.setStyle(Qt.PenStyle.DotLine)
        painter.setPen(pen)
        painter.drawLine(QLineF(0, x_axis, width, x_axis))

        # Draw gray transparent overlay if track is disabled.
        if self.disabled:
            painter.fillPath(path, self.DISABLED_WAVEFORM_OVERLAY_COLOR)

    def enable(self) -> None:
        self.disabled = False
        self.update()

    def disable(self) -> None:
        self.disabled = True
        self.update()
